/*
  Macro           eHelp_CreateIndex
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all TSE variants   (Intended, not tested yet)
  Version         v0.22   23 Mar 2023


  This tool (re)creates a tsehelp_index.ehelp file based on optional parameters
  and based on those non-index .ehelp files, that are configured in eHelp.

  Here "non-index files" case-insensitively means files without "index" in
  their names.

  For better and for worse tsehelp_index.ehelp indexes ALL topics and subtopics.



  INSTALLATION

  This tool will by installed by eHelp if you follow eHelp's installation
  instructions.

  Alternatively you can install it yourself by placing this file in TSE's "mac"
  directory, and by compiling it there, for example by opening the file in TSE
  and applying the Macro -> Compile menu.


  USAGE

  Execute the tool as a macro without parameters to get an interactive input
  menu and to get interactive output.

  Execute the tool with one of these parameters to skip interactivity:
    "outdated"  Only (re)create tsehelp_index.ehelp if it does not exist or if
                it has a date-time that does not equal the date-time of the
                newest configured non-index .ehelp file.
    "always"    Always (re)create tsehelp_index.ehelp.

    For example:
      ExecMacro('eHelp_CreateIndex outdated')

  User results:
  - If there was no error then the tsehelp_index.ehelp file is (re)created,
    and its date-time is set to the date-time of the newest configured
    non-index .ehelp file.

  Technical results:
  - TSE's "MacroCmdLine" variable is set to "1" if eveything went OK,
    and to "0" if there was an error.
  - If
    - a tsehelp_index.ehelp was created without an error,
    - and tsehelp_index.ehelp does not exist yet in eHelp's configured .ehelp
      files,
    then eHelp_CreateIndex
    - adds tsehelp_index.ehelp to the end of the configured .ehelp files,
      where eHelp will later position it to the correct default position,
    - and sets a buffer integer so eHelp knows to do that ASAP.


  KNOWN POTENTIAL ISSUE

  I do not know yet, whether redirecting the topic in a subtopic-reference
  should work.


  HISTORY

  v0.1    9 Jan 2023
    Initial release.

  v0.2   10 Jan 2023
    Added lines between topics with a different first character:
    alphabetic first characters get a proper header line,
    other first characters just get an empty line.

  v0.18   8 Mar 2023
    All eHelp* macros now need to have the same version number.
      This is enforced to simplify the maintenance of the inter-compatibility
      between these macros.
    Fixed responsabilaties and logic order between eHelp and eHelp_CreateIndex
    for the default configuration of .ehelp files and creating this index.
    Fixed generating an empty index when installing eHelp.
    Implemented differentiating between index and non-index .ehelp files
    based on (not) having "index" in their name.
    Added interactivity if no parameters are supplied.
*/





// Start of compatibility restrictions and mitigations


#ifdef LINUX
  #define WIN32 FALSE
#endif


#ifdef WIN32
#else
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif


#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#ifdef LINUX
  string SLASH [1] = '/'
#else
  string SLASH [1] = '\'
#endif


// End of compatibility restrictions and mitigations





// Global constants and semi-constants
#define CHANGECURRFILENAME_OPTIONS               _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
integer EHELP_FILES_ID                         = 0
string  HELP_TOOL                          [5] = 'eHelp'
string  INDEXER_ADDED_INDEX_VAR [MAXSTRINGLEN] = ''
string  INDEX_EHELP_FILE_NAME             [19] = 'tsehelp_index.ehelp'

#define INDEX_STATUS_INCORRECT                   4
#define INDEX_STATUS_NOT_EXISTS                  1
#define INDEX_STATUS_OK                          5
#define INDEX_STATUS_OUTDATED                    3
#define INDEX_STATUS_UNDETERMINED                0

#define INVALID_HANDLE_VALUE                     -1

string  MACRO_NAME              [MAXSTRINGLEN] = ''
string  MACRO_VERSION                      [5] = 'v0.22'

#define NEGATIVE_MAXSTRINGLEN                    MAXSTRINGLEN * -1

#define SYSTEM_HELP_CONTROL_ID                   9
#define SYSTEM_HELP_HEADERS_ID                   7
#define SYSTEM_HELP_INDEX_ID                     10
#define SYSTEM_HELP_TOPIC_ID                     8

#define UPDATE_ALWAYS                            1
#define UPDATE_NOT                               3
#define UPDATE_OUTDATED                          2
#define UPDATE_UNDETERMINED                      0


// Global variables
integer interactive_mode = FALSE
integer update_type      = UPDATE_UNDETERMINED


// Windows DLLs

#if WIN32

  dll "<Kernel32.dll>"

    integer proc CloseHandle(
      integer handle
      ): "CloseHandle"

    integer proc CreateFile(
      string  lpFileName:cstrval,
      integer dwDesiredAccess,
      integer dwShareMode,
      integer lpSecurityAttributes,
      integer dwCreationDisposition,
      integer dwFlagsAndAttributes,
      integer hTemplateFile
      ): "CreateFileA"

    integer proc DosDateTimeToFileTime(
      integer wFatDate:word,
      integer wFatTime:word,
      string  lpFileTime:StrPtr
      ): "DosDateTimeToFileTime"

    integer proc GetLastError(
      ): "GetLastError"

    integer proc LocalFileTimeToFileTime(
      string in_time:StrPtr,
      string out_time:StrPtr
      ): "LocalFileTimeToFileTime"

    integer proc SetFileTime(
      integer handle,
      integer lpCreationTime,
      integer lpLastAccessTime,
      string  lpLastWriteTime:StrPtr
      ): "SetFileTime"

  end
#endif


// Local procedures

#if WIN32
  /*
    This is an IMHO optimized version of the "f" macro's LoSetDateTime() proc.
    It has a higher-level date-time input and better documentation.
  */
  integer proc set_windows_file_date_time(string fn, string yyyy_mm_dd_hh_mm_ss)
    integer ff_date             = 0
    integer ff_time             = 0
    integer handle              = INVALID_HANDLE_VALUE
    string  local_file_time [8] = "        "  // Needs to be at least 64 bits long.
    integer ok                  = FALSE
    string  utc_file_time   [8] = "        "  // Needs to be at least 64 bits long.

    //  Bitfields for file date, bits counted right-to-left, starting at 0:
    //    15-9   year-1980
    //     8-5   month
    //     4-0   day
    ff_date = ( ((Val(yyyy_mm_dd_hh_mm_ss[1:4]) - 1980) shl  9) |
                 (Val(yyyy_mm_dd_hh_mm_ss[6:2])         shl  5) |
                  Val(yyyy_mm_dd_hh_mm_ss[9:2])                 )

    //  Bitfields for file time, bits counted right-to-left, starting at 0:
    //    15-11  hours (0-23)
    //    10-5   minutes
    //     4-0   seconds/2
    ff_time = (  (Val(yyyy_mm_dd_hh_mm_ss[12:2])         shl 11) |
                 (Val(yyyy_mm_dd_hh_mm_ss[15:2])         shl  5) |
                 (Val(yyyy_mm_dd_hh_mm_ss[18:2]) /   2)          )

    if not DosDateTimeToFileTime(ff_date, ff_time, local_file_time)
      Warn(MACRO_NAME, ': Error ON DosDateTimeToFileTime()')
      return(FALSE)
    endif

    // The 2nd parameter, "dwDesiredAccess" in the Microsoft documentation,
    // has the value GENERIC_WRITE, which is 0x40000000.
    // The 5th parameter, "dwCreationDisposition" in the Microsoft documentation,
    // has the value OPEN_EXISTING, which is 3.
    handle = CreateFile(fn, 0x40000000, 0, 0, 3, 0, 0)

    if handle == INVALID_HANDLE_VALUE
      Warn(MACRO_NAME, ': Error ', GetLastError(), ' opening file ', fn)
      return(FALSE)
    endif

    // The handle is open, so from here on we cannot simply return().

    ok = LocalFileTimeToFileTime(local_file_time, utc_file_time)
    if not ok
      Warn(MACRO_NAME, ': Error ', GetLastError(), ' converting local to UTC time.')
    endif

    // The two 0's are lpCreationTime and lpLastAccessTime,
    // but to pass 0 they have to have been declared as integers,
    // as was done in their DLL declaration above.
    ok = ok and SetFileTime(handle, 0, 0, utc_file_time)
    if not ok
      Warn(MACRO_NAME, ': Error ', GetLastError(), ' setting time, handle:', handle)
    endif

    CloseHandle(handle)

    return(ok)
  end set_windows_file_date_time
#endif


#ifdef LINUX
  integer proc set_linux_file_date_time(string fn, string yyyy_mm_dd_hh_mm_ss)
    string  cmd  [MAXSTRINGLEN] = ''
    integer ok                  = FALSE
    integer command_return_code = 0
    integer dos_errorlevel      = 0
    integer dos_io_result       = 0
    integer dos_result          = 0

    cmd = Format('touch -t ',
                 yyyy_mm_dd_hh_mm_ss[ 1:4],
                 yyyy_mm_dd_hh_mm_ss[ 6:2],
                 yyyy_mm_dd_hh_mm_ss[ 9:2],
                 yyyy_mm_dd_hh_mm_ss[12:2],
                 yyyy_mm_dd_hh_mm_ss[15:2],
                 '.',
                 yyyy_mm_dd_hh_mm_ss[18:2],
                 ' ',
                 fn)

    dos_result          = Dos(cmd, DOS_SYNC_CALL_FLAGS)
    dos_io_result       = DosIOResult()
    command_return_code = HiByte(dos_io_result)
    dos_errorlevel      = LoByte(dos_io_result)

    if      dos_result
    and not command_return_code
    and not dos_errorlevel
      ok = TRUE
    else
      Warn(MACRO_NAME, ' error executing:'; cmd)
    endif

    return(ok)
  end set_linux_file_date_time
#endif


integer proc set_file_date_time(string fn, string yyyy_mm_dd_hh_mm_ss)
  integer ok = FALSE

  #if WIN32
    ok = set_windows_file_date_time(fn, yyyy_mm_dd_hh_mm_ss)
  #else
    ok = set_linux_file_date_time(fn, yyyy_mm_dd_hh_mm_ss)
  #endif
  return(ok)
end set_file_date_time




/*
  search_path works like SearchPath, but without looking in the current
  directory too if subdir is supplied.
  For example, if subdir is supplied, and paths is "path1;path2", then these
  directories are searched in this order by search_path:
    path1
    path1\subdir
    path2
    path2\subdir
    LoadDir()
    LoadDir()\subdir
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)\subdir

  search_path_helper() should not be called directly.
  It is a helper proc for search_path().
*/
string proc search_path_helper(string filename, string path , string subdir)
  string result [MAXSTRINGLEN] = ''

  if FileExists(path + filename)
    result = path + filename
  elseif subdir <> ''
  and    FileExists(path + subdir + SLASH + filename)
    result = path + subdir + SLASH + filename
  endif

  return(result)
end search_path_helper


string proc search_path(string filename, string paths, string subdir)
  integer i                          = 0
  string  path        [MAXSTRINGLEN] = ''
  string  program_dir [MAXSTRINGLEN] = ''
  string  result      [MAXSTRINGLEN] = ''

  for i = 1 to NumTokens(paths, ';')
    path   = GetToken(paths, ';', i)
    result = search_path_helper(filename, path + SLASH, subdir)
    if result <> ''
      break
    endif
  endfor

  if result == ''
    path   = LoadDir()
    result = search_path_helper(filename, path, subdir)
  endif

  if result == ''
    program_dir = SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    if program_dir <> path
      result = search_path_helper(filename, program_dir, subdir)
    endif
  endif

  return(result)
end search_path


string proc get_file_date_time(string file_fqn)
  string file_date_time [19] = ''
  if FindThisFile(file_fqn)
      file_date_time = FFDateStr() + ' ' + FFTimeStr()
  endif
  return(file_date_time)
end get_file_date_time


/*
  Give a warning if WriteProfileStr() fails.
*/
integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer result = WriteProfileStr(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn('Failed to write'; MACRO_NAME; 'configuration to .ini file.')
  endif
  return(result)
end write_profile_str


integer proc find_profile_value(    string  index_file_fqn,
                                var integer highest_profile_index)
  string  dummy_index    [MAXSTRINGLEN] = ''
  string  ehelp_file_fqn [MAXSTRINGLEN] = ''
  integer result                        = FALSE

  highest_profile_index = 0

  LoadProfileSection(HELP_TOOL + ':Files')
  while GetNextProfileItem(dummy_index, ehelp_file_fqn)
    if EquiStr(ehelp_file_fqn, index_file_fqn)
      result = TRUE
    endif
    if Val(dummy_index) > highest_profile_index
      highest_profile_index = Val(dummy_index)
    endif
  endwhile
  return(result)
end find_profile_value


integer proc create_index_file(string index_file_fqn,
                               string max_non_index_file_date_time)
  integer ehelp_buffer_id                   = 0
  string  ehelp_file_fqn     [MAXSTRINGLEN] = ''
  string  first_character_next          [1] = ''
  string  first_character_prev          [1] = ''
  integer highest_profile_index             = 0
  integer index_buffer_id                   = 0
  string  link               [MAXSTRINGLEN] = ''
  string  link_address       [MAXSTRINGLEN] = ''
  string  link_comment       [MAXSTRINGLEN] = ''
  string  link_display       [MAXSTRINGLEN] = ''
  string  link_sort_key      [MAXSTRINGLEN] = ''
  string  link_sort_key_prev [MAXSTRINGLEN] = ''
  string  link_subtopic      [MAXSTRINGLEN] = ''
  string  link_topic         [MAXSTRINGLEN] = ''
  integer ok                                = TRUE
  string  old_MacroCmdLine   [MAXSTRINGLEN] = ''
  integer old_MsgLevel                      = 0
  string  redirect           [MAXSTRINGLEN] = ''
  string  redirect_from      [MAXSTRINGLEN] = ''
  string  redirect_to        [MAXSTRINGLEN] = ''
  integer redirects_line_from               = 0
  integer redirects_line_to                 = 0
  integer topics_line_from                  = 0
  integer topics_line_to                    = 0

  PushLocation()
  index_buffer_id = CreateTempBuffer()
  ChangeCurrFilename(index_file_fqn, CHANGECURRFILENAME_OPTIONS)
  ehelp_buffer_id = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':ehelp_file', CHANGECURRFILENAME_OPTIONS)

  GotoBufferId(ehelp_files_id)
  if NumLines()
    BegFile()
    repeat
      ehelp_file_fqn = GetText(1, MAXSTRINGLEN)

      GotoBufferId(ehelp_buffer_id)
      if LoadBuffer(ehelp_file_fqn)
        redirects_line_from = 0
        redirects_line_to   = 0

        if lFind('^.*</REDIRECTS>.*$', 'bigx')
          redirects_line_to = CurrLine()
          if lFind('^.*<REDIRECTS>.*$', 'bix')
            redirects_line_from = CurrLine()
          endif
        endif

        // xx   To be revisited:
        // xx   These links might be imsufficient, depending on if/how
        // xx   redirecting the topic of a subtopic works.
        if  redirects_line_from <> 0
        and redirects_line_to   <> 0
          while lFind('^ *<REDIRECT>{.+}</REDIRECT> *$', 'ix+')
          and   CurrLine() < redirects_line_to
            redirect = GetFoundText(1)
            if Pos('^', redirect)
              redirect_from = redirect[1: Pos('^', redirect) - 1]
              redirect_to   = redirect[Pos('^', redirect) + 1: MAXSTRINGLEN]
              if Pos(';', redirect_to)
                link_topic    = redirect_to[1: Pos(';', redirect_to) - 1]
                link_subtopic = redirect_to[Pos(';', redirect_to) + 1: MAXSTRINGLEN]
                if EquiStr(redirect_from, link_subtopic)
                  link         = redirect_to + '^' + redirect_from
                  link_comment = ' (' + link_topic + ')'
                else
                  link         = redirect_from
                  link_comment = ' (' + redirect_to + ')'
                endif
              else
                link         = redirect_from
                link_comment = ' (' + redirect_to + ')'
              endif
            endif
            AddLine('<LINK>' + link + '</LINK>' + link_comment, index_buffer_id)
          endwhile
        endif

        topics_line_from = 0
        topics_line_to   = 0

        if lFind('^.*</TOPICS>.*$', 'bigx')
          topics_line_to = CurrLine()
          if lFind('^.*<TOPICS>.*$', 'bix')
            topics_line_from = CurrLine()
          endif
        endif

        if  topics_line_from <> 0
        and topics_line_to   <> 0
          GotoLine(topics_line_from)
          while lFind('^ *<TOPIC>{.+}</TOPIC> *$', 'ix+')
          and   CurrLine() < topics_line_to
            link_topic = GetFoundText(1)
            link       = link_topic
            AddLine('<LINK>' + link + '</LINK>', index_buffer_id)
          endwhile

          GotoLine(topics_line_from)
          while lFind('<SUBTOPIC>{.+}</SUBTOPIC>', 'ix+')
          and   CurrLine() < topics_line_to
            link_subtopic = GetFoundText(1)
            link_display  = link_subtopic
            PushLocation()
            if lFind('^ *<TOPIC>{.+}</TOPIC> *$', 'bix+')
              link_topic   = GetFoundText(1)
              link_address = link_topic   + ';' + link_subtopic
              link         = link_address + '^' + link_display
              AddLine('<LINK>' + link + '</LINK> (' + link_topic + ')',
                      index_buffer_id)
            else
              Warn(MACRO_NAME, ': subtopic without a topic: "',
                   link_subtopic, '" in "', CurrFilename(), '".')
            endif
            PopLocation()
          endwhile
        endif
      endif

      GotoBufferId(ehelp_files_id)
    until not Down()

    // Sort on [subtopic + ';' + ] topic and remove duplicates.
    GotoBufferId(index_buffer_id)

    //   Step 1: Temporarily prefix each line with its sort key.
    BegFile()
    repeat
      link_sort_key = ''
      if lFind('^ *<LINK>{[~\^]+}</LINK>'    , 'cgix')
      or lFind('^ *<LINK>{[~\^]+}\^.*</LINK>', 'cgix')
        link_address = GetFoundText(1)
        if Pos(';', link_address)
          link_topic    = link_address[1: Pos(';', link_address) - 1]
          link_subtopic = link_address[Pos(';', link_address) + 1: MAXSTRINGLEN]
          link_sort_key = link_subtopic + ';' + link_topic
        else
          link_sort_key = link_address
        endif
        BegLine()
        InsertText(Format(link_sort_key: NEGATIVE_MAXSTRINGLEN), _INSERT_)
      else
        InsertText(Format('!'          : NEGATIVE_MAXSTRINGLEN), _INSERT_)
      endif
    until not Down()

    //   Step 2: Sort the buffer on the sort key.
    PushBlock()
    MarkLine(1, NumLines())
    old_MacroCmdLine = Query(MacroCmdLine)
    old_MsgLevel     = Set(MsgLevel, _WARNINGS_ONLY_)
    ExecMacro('sort -i') // Cannot use "-k" because of backwards incompatibility.
    Set(MsgLevel    , old_MsgLevel)
    Set(MacroCmdLine, old_MacroCmdLine)
    PopBlock()

    //   Step 3: Remove duplicate sort keys.
    BegFile()
    link_sort_key_prev = GetText(1, MAXSTRINGLEN)
    while Down()
      link_sort_key = GetText(1, MAXSTRINGLEN)
      if EquiStr(link_sort_key, link_sort_key_prev)
        KillLine()
        Up()
      else
        link_sort_key_prev = link_sort_key
      endif
    endwhile

    //   Step 4: Remove the sort key.
    PushBlock()
    MarkColumn(1, 1, NumLines(), MAXSTRINGLEN)
    KillBlock()
    PopBlock()

    // Insert text lines to separate lines with the same first character in
    // their link's display value.
    BegFile()
    if lFind('^ *<LINK>{[~\^]+}</LINK>'        , 'cgix')
    or lFind('^ *<LINK>[~\^]+\^{[~\^]+}</LINK>', 'cgix')
      first_character_prev = Upper(SubStr(GetFoundText(1), 1, 1))
    else
      first_character_prev = ''
    endif
    while Down()
      if lFind('^ *<LINK>{[~\^]+}</LINK>'        , 'cgix')
      or lFind('^ *<LINK>[~\^]+\^{[~\^]+}</LINK>', 'cgix')
        first_character_next = Upper(SubStr(GetFoundText(1), 1, 1))
      else
        first_character_next = ''
      endif
      if first_character_next <> first_character_prev
        if isAlpha(first_character_next)
          InsertLine('')
          AddLine('<BOLD>--- ' + first_character_next + ' ---</BOLD>')
          AddLine('')
        else
          InsertLine('')
        endif
        Down()
      endif
      first_character_prev = first_character_next
    endwhile
  endif

  BegFile()
  InsertLine('<EHELP>')
  AddLine('')
  AddLine('<META>')
  AddLine('  <INDEXER_TOOL>'      + MACRO_NAME    + '</GENERATOR_TOOL>'   )
  AddLine('  <INDEXER_VERSION>'   + MACRO_VERSION + '</GENERATOR_VERSION>')
  AddLine('  <INDEXED_DATE_TIME>' + GetDateStr() + ' ' + GetTimeStr() +
                                                    '</INDEXED_DATE_TIME>')
  AddLine('  <TEXT>')
  AddLine('    Do not modify this tsehelp_index.ehelp file manually!')
  AddLine('    It will be overwritten when one of the configured .ehelp files changes.')
  AddLine('  </TEXT>')
  AddLine('</META>')
  AddLine('')
  AddLine('<TOPICS>')
  AddLine('<TOPIC>Index</TOPIC>')
  AddLine('<HELP>')

  EndFile()
  AddLine('</HELP>')
  AddLine('</TOPICS>')
  AddLine('</EHELP>')

  if SaveAs(index_file_fqn, _DONT_PROMPT_|_OVERWRITE_)
    if not find_profile_value(index_file_fqn, highest_profile_index)
      // This writes the new profile entry to the end of the profile section.
      // (Side note: the entry name has no influence on that.)
      // eHelp will move the profile entry to its correct default position.
      ok = write_profile_str(HELP_TOOL + ':Files',
                             Str(highest_profile_index + 1),
                             index_file_fqn)
      if ok
        SetBufferInt(INDEXER_ADDED_INDEX_VAR, TRUE, SYSTEM_HELP_TOPIC_ID)
      endif
    endif

    set_file_date_time(index_file_fqn, max_non_index_file_date_time)
  endif

  PopLocation()
  AbandonFile(ehelp_buffer_id)
  AbandonFile(index_buffer_id)

  return(ok)
end create_index_file


//  Determines the global variables
//    index_file_fqn        (future name if no index yet)
//    index_file_date_time  (empty       if no index yet)
//    max_non_index_file_date_time
//  and fills the buffer
//    EHELP_FILES_ID
//  with the configured non-index filenames.
proc collect_resource_data(var string index_file_fqn,
                           var string index_file_date_time,
                           var string max_non_index_file_date_time)
  string  dummy_name                 [1] = ''
  string  ehelp_file_date_time      [19] = ''
  string  ehelp_file_fqn  [MAXSTRINGLEN] = ''
  string  prog_dir        [MAXSTRINGLEN] = ''

  LoadProfileSection(HELP_TOOL + ':Files')
  while GetNextProfileItem(dummy_name, ehelp_file_fqn)
    if EquiStr(SplitPath(ehelp_file_fqn, _NAME_|_EXT_), INDEX_EHELP_FILE_NAME)
      index_file_fqn       = ehelp_file_fqn
      index_file_date_time = get_file_date_time(ehelp_file_fqn)
    elseif not Pos('index', Lower(SplitPath(ehelp_file_fqn, _NAME_)))
      // This branch is skipped for any non-standard index-named files.
      AddLine(ehelp_file_fqn, EHELP_FILES_ID)
      ehelp_file_date_time = get_file_date_time(ehelp_file_fqn)
      if ehelp_file_date_time > max_non_index_file_date_time
        max_non_index_file_date_time = ehelp_file_date_time
      endif
    endif
  endwhile

  if index_file_fqn == ''
    index_file_fqn = search_path(INDEX_EHELP_FILE_NAME, Query(TSEPath), 'help')
    if index_file_fqn == ''
      if FileExists(LoadDir() + 'help') & _DIRECTORY_
        index_file_fqn = LoadDir() + 'help\' + INDEX_EHELP_FILE_NAME
      else
        if FileExists(LoadDir()) & _DIRECTORY_
          MkDir(LoadDir() + 'help')
          index_file_fqn = LoadDir() + 'help\' + INDEX_EHELP_FILE_NAME
        else
          prog_dir = SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
          if FileExists(prog_dir + 'help') & _DIRECTORY_
            index_file_fqn = prog_dir + 'help\' + INDEX_EHELP_FILE_NAME
          endif
        endif
      endif
    endif
  endif
end collect_resource_data


integer proc get_index_status(string index_file_fqn,
                              string index_file_date_time,
                              string max_non_index_file_date_time)
  integer index_status = INDEX_STATUS_UNDETERMINED

  if FileExists(index_file_fqn)
    if     index_file_date_time <  max_non_index_file_date_time
      index_status = INDEX_STATUS_OUTDATED
    elseif index_file_date_time <> max_non_index_file_date_time
      index_status = INDEX_STATUS_INCORRECT
    else
      index_status = INDEX_STATUS_OK
    endif
  else
    index_status = INDEX_STATUS_NOT_EXISTS
  endif

  return(index_status)
end get_index_status


proc set_integer_var(var integer integer_reference, integer new_integer_value)
  integer_reference = new_integer_value
end set_integer_var


menu update_type_menu()
  title = 'Create new tsehelp_index.ehelp file from configured .ehelp files?'
  history

  '&Only if outdated', set_integer_var(update_type, UPDATE_OUTDATED)
  '&Always'          , set_integer_var(update_type, UPDATE_ALWAYS  )
  '&No'              , set_integer_var(update_type, UPDATE_NOT     )
end update_type_menu


integer proc check_and_possibly_create_index()
  integer i                                 = 0
  string  index_file_date_time         [19] = ''
  string  index_file_fqn     [MAXSTRINGLEN] = ''
  integer index_status                      = INDEX_STATUS_UNDETERMINED
  string  macro_command_line [MAXSTRINGLEN] = Query(MacroCmdLine)
  string  max_non_index_file_date_time [19] = ''
  integer ok                                = TRUE

  for i = 1 to NumTokens(macro_command_line, ' ')
    case Lower(GetToken(macro_command_line, ' ', i))
      when 'always'
        update_type = UPDATE_ALWAYS
      when 'outdated'
        update_type = UPDATE_OUTDATED
      otherwise
        Warn(MACRO_NAME; ' ERROR: Illegal parameter "', GetToken(macro_command_line, ' ', i), '".')
        ok = FALSE
        break
    endcase
  endfor

  if  ok
  and update_type == UPDATE_UNDETERMINED
    interactive_mode = TRUE
    update_type_menu()
  endif

  if ok
    collect_resource_data(index_file_fqn,
                          index_file_date_time,
                          max_non_index_file_date_time)

    index_status = get_index_status(index_file_fqn,
                                    index_file_date_time,
                                    max_non_index_file_date_time)
    if index_file_fqn <> ''
      if     update_type == UPDATE_ALWAYS
      or (   update_type == UPDATE_OUTDATED
         and ( index_status in INDEX_STATUS_INCORRECT,
                               INDEX_STATUS_NOT_EXISTS,
                               INDEX_STATUS_OUTDATED))

        ok = create_index_file(index_file_fqn,
                               max_non_index_file_date_time)
      endif
    endif
  endif

  if  ok
  and interactive_mode
    case update_type
      when UPDATE_NOT
        Warn('No action.')
      when UPDATE_ALWAYS
        if index_status == INDEX_STATUS_NOT_EXISTS
          Warn('tsehelp_index.ehelp has been created.')
        else
          Warn('tsehelp_index.ehelp has been recreated.')
        endif
      when UPDATE_OUTDATED
        case index_status
          when INDEX_STATUS_INCORRECT
            Warn('tsehelp_index.ehelp was incorrect and has been recreated.')
          when INDEX_STATUS_NOT_EXISTS
            Warn('tsehelp_index.ehelp did not exist and has been created.')
          when INDEX_STATUS_OUTDATED
            Warn('tsehelp_index.ehelp was outdated and has been recreated.')
          otherwise
            Warn('tsehelp_index.ehelp was not outdated and has not been recreated.')
        endcase
      otherwise
        Warn('You escaped!')
    endcase
  endif

  return(ok)
end check_and_possibly_create_index


proc WhenPurged()
  AbandonFile(EHELP_FILES_ID)
end WhenPurged


proc WhenLoaded()
  MACRO_NAME              = SplitPath(CurrMacroFilename(), _NAME_)
  INDEXER_ADDED_INDEX_VAR = MACRO_NAME + ':AddedIndex'

  DelBufferVar(INDEXER_ADDED_INDEX_VAR, SYSTEM_HELP_TOPIC_ID)

  PushLocation()
  EHELP_FILES_ID = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':configured_non_index_ehelp_files',
                     CHANGECURRFILENAME_OPTIONS)
  PopLocation()
end WhenLoaded


proc Main()
  integer ok                        = TRUE
  integer old_DateFormat            = Set(DateFormat           , 6)
  integer old_DateSeparator         = Set(DateSeparator        , Asc('-'))
  integer old_Insert                = Set(Insert               , ON)
  integer old_InsertLineBlocksAbove = Set(InsertLineBlocksAbove, FALSE)
  integer old_RemoveTrailingWhite   = Set(RemoveTrailingWhite  , ON)
  integer old_TimeFormat            = Set(TimeFormat           , 1)
  integer old_TimeSeparator         = Set(TimeSeparator        , Asc(':'))
  integer old_UnMarkAfterPaste      = Set(UnMarkAfterPaste     , ON)

  ok = check_and_possibly_create_index()

  if ok
    // Set buffer string to index_file_name?
  endif

  Set(MacroCmdLine, Str(ok))
  SetBufferStr(MACRO_NAME + ':Version', MACRO_VERSION, SYSTEM_HELP_TOPIC_ID)

  Set(DateFormat           , old_DateFormat)
  Set(DateSeparator        , old_DateSeparator)
  Set(Insert               , old_Insert)
  Set(InsertLineBlocksAbove, old_InsertLineBlocksAbove)
  Set(RemoveTrailingWhite  , old_RemoveTrailingWhite)
  Set(TimeFormat           , old_TimeFormat)
  Set(TimeSeparator        , old_TimeSeparator)
  Set(UnMarkAfterPaste     , old_UnMarkAfterPaste)

  PurgeMacro(MACRO_NAME)
end Main
